"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const events_1 = require("events");
const constants_1 = require("./constants");
const errors_1 = require("./errors");
const events_2 = require("./events");
const peer_book_1 = require("./peer_book");
const peer_pool_1 = require("./peer_pool");
const peer_server_1 = require("./peer_server/");
const utils_1 = require("./utils");
const createPeerPoolConfig = (config, peerBook) => ({
    connectTimeout: config.connectTimeout,
    ackTimeout: config.ackTimeout,
    wsMaxPayload: config.wsMaxPayload
        ? config.wsMaxPayload
        : constants_1.DEFAULT_WS_MAX_PAYLOAD,
    peerSelectionForSend: config.peerSelectionForSend
        ? config.peerSelectionForSend
        : utils_1.selectPeersForSend,
    peerSelectionForRequest: config.peerSelectionForRequest
        ? config.peerSelectionForRequest
        : utils_1.selectPeersForRequest,
    peerSelectionForConnection: config.peerSelectionForConnection
        ? config.peerSelectionForConnection
        : utils_1.selectPeersForConnection,
    sendPeerLimit: config.sendPeerLimit === undefined
        ? constants_1.DEFAULT_SEND_PEER_LIMIT
        : config.sendPeerLimit,
    peerBanTime: config.peerBanTime ? config.peerBanTime : constants_1.DEFAULT_BAN_TIME,
    maxOutboundConnections: config.maxOutboundConnections === undefined
        ? constants_1.DEFAULT_MAX_OUTBOUND_CONNECTIONS
        : config.maxOutboundConnections,
    maxInboundConnections: config.maxInboundConnections === undefined
        ? constants_1.DEFAULT_MAX_INBOUND_CONNECTIONS
        : config.maxInboundConnections,
    maxPeerDiscoveryResponseLength: config.maxPeerDiscoveryResponseLength === undefined
        ? constants_1.DEFAULT_MAX_PEER_DISCOVERY_RESPONSE_LENGTH
        : config.maxPeerDiscoveryResponseLength,
    maxPeerInfoSize: config.maxPeerInfoSize
        ? config.maxPeerInfoSize
        : constants_1.DEFAULT_MAX_PEER_INFO_SIZE,
    outboundShuffleInterval: config.outboundShuffleInterval
        ? config.outboundShuffleInterval
        : constants_1.DEFAULT_OUTBOUND_SHUFFLE_INTERVAL,
    netgroupProtectionRatio: typeof config.netgroupProtectionRatio === 'number'
        ? config.netgroupProtectionRatio
        : constants_1.DEFAULT_PEER_PROTECTION_FOR_NETGROUP,
    latencyProtectionRatio: typeof config.latencyProtectionRatio === 'number'
        ? config.latencyProtectionRatio
        : constants_1.DEFAULT_PEER_PROTECTION_FOR_LATENCY,
    productivityProtectionRatio: typeof config.productivityProtectionRatio === 'number'
        ? config.productivityProtectionRatio
        : constants_1.DEFAULT_PEER_PROTECTION_FOR_USEFULNESS,
    longevityProtectionRatio: typeof config.longevityProtectionRatio === 'number'
        ? config.longevityProtectionRatio
        : constants_1.DEFAULT_PEER_PROTECTION_FOR_LONGEVITY,
    wsMaxMessageRate: typeof config.wsMaxMessageRate === 'number'
        ? config.wsMaxMessageRate
        : constants_1.DEFAULT_WS_MAX_MESSAGE_RATE,
    wsMaxMessageRatePenalty: typeof config.wsMaxMessageRatePenalty === 'number'
        ? config.wsMaxMessageRatePenalty
        : constants_1.DEFAULT_WS_MAX_MESSAGE_RATE_PENALTY,
    rateCalculationInterval: typeof config.rateCalculationInterval === 'number'
        ? config.rateCalculationInterval
        : constants_1.DEFAULT_RATE_CALCULATION_INTERVAL,
    secret: config.secret ? config.secret : constants_1.DEFAULT_RANDOM_SECRET,
    peerBook,
});
class P2P extends events_1.EventEmitter {
    constructor(config) {
        super();
        this._secret = config.secret ? config.secret : constants_1.DEFAULT_RANDOM_SECRET;
        this._sanitizedPeerLists = utils_1.sanitizePeerLists({
            seedPeers: config.seedPeers
                ? config.seedPeers.map(utils_1.sanitizeInitialPeerInfo)
                : [],
            blacklistedIPs: config.blacklistedIPs ? config.blacklistedIPs : [],
            fixedPeers: config.fixedPeers
                ? config.fixedPeers.map(utils_1.sanitizeInitialPeerInfo)
                : [],
            whitelisted: config.whitelistedPeers
                ? config.whitelistedPeers.map(utils_1.sanitizeInitialPeerInfo)
                : [],
            previousPeers: config.previousPeers
                ? config.previousPeers.map(utils_1.sanitizeInitialPeerInfo)
                : [],
        }, {
            peerId: utils_1.constructPeerId(config.hostIp || constants_1.DEFAULT_NODE_HOST_IP, config.nodeInfo.wsPort),
            ipAddress: config.hostIp || constants_1.DEFAULT_NODE_HOST_IP,
            wsPort: config.nodeInfo.wsPort,
        }, this._secret);
        this._config = config;
        this._isActive = false;
        this._hasConnected = false;
        this._peerBook = new peer_book_1.PeerBook({
            sanitizedPeerLists: this._sanitizedPeerLists,
            secret: this._secret,
        });
        this._handlePeerPoolRPC = (request) => {
            switch (request.procedure) {
                case events_2.REMOTE_EVENT_RPC_GET_PEERS_LIST:
                    this._handleGetPeersRequest(request);
                    break;
                case events_2.REMOTE_EVENT_RPC_GET_NODE_INFO:
                    this._handleGetNodeInfo(request);
                    break;
                default:
            }
            this.emit(events_2.EVENT_REQUEST_RECEIVED, request);
        };
        this._handlePeerPoolMessage = (message) => {
            this.emit(events_2.EVENT_MESSAGE_RECEIVED, message);
        };
        this._handleOutboundPeerConnect = (peerInfo) => {
            if (!this._peerBook.hasPeer(peerInfo)) {
                this._peerBook.addPeer(peerInfo);
            }
            this._peerBook.upgradePeer(peerInfo);
            this.emit(events_2.EVENT_CONNECT_OUTBOUND, peerInfo);
            if (this._isNetworkReady()) {
                this.emit(events_2.EVENT_NETWORK_READY);
            }
        };
        this._handleOutboundPeerConnectAbort = (peerInfo) => {
            if (this._peerBook.hasPeer(peerInfo)) {
                this._peerBook.downgradePeer(peerInfo);
            }
            this.emit(events_2.EVENT_CONNECT_ABORT_OUTBOUND, peerInfo);
        };
        this._handlePeerCloseOutbound = (closePacket) => {
            const { peerInfo } = closePacket;
            if (this._peerBook.getPeer(closePacket.peerInfo)) {
                const updatedPeer = {
                    ...peerInfo,
                    internalState: utils_1.assignInternalInfo(peerInfo, this._secret),
                };
                this._peerBook.updatePeer(updatedPeer);
            }
            this.emit(events_2.EVENT_CLOSE_OUTBOUND, closePacket);
        };
        this._handlePeerCloseInbound = (closePacket) => {
            const { peerInfo } = closePacket;
            if (this._peerBook.getPeer(closePacket.peerInfo)) {
                const updatedPeer = {
                    ...peerInfo,
                    internalState: utils_1.assignInternalInfo(peerInfo, this._secret),
                };
                this._peerBook.updatePeer(updatedPeer);
            }
            this.emit(events_2.EVENT_CLOSE_INBOUND, closePacket);
        };
        this._handleFailedInboundPeerConnect = (err) => {
            this.emit(events_2.EVENT_FAILED_TO_ADD_INBOUND_PEER, err);
        };
        this._handleInboundPeerConnect = (incomingPeerConnection) => {
            try {
                this._peerPool.addInboundPeer(incomingPeerConnection.peerInfo, incomingPeerConnection.socket);
                if (!this._peerBook.hasPeer(incomingPeerConnection.peerInfo)) {
                    this._peerBook.addPeer({
                        ...incomingPeerConnection.peerInfo,
                        sourceAddress: incomingPeerConnection.socket.remoteAddress,
                    });
                }
                this.emit(events_2.EVENT_NEW_INBOUND_PEER, incomingPeerConnection.peerInfo);
                return;
            }
            catch (err) {
                if (err instanceof errors_1.PeerInboundDuplicateConnectionError) {
                    incomingPeerConnection.socket.disconnect(constants_1.DUPLICATE_CONNECTION, constants_1.DUPLICATE_CONNECTION_REASON);
                    this.emit(events_2.EVENT_FAILED_TO_ADD_INBOUND_PEER, err);
                    return;
                }
                incomingPeerConnection.socket.disconnect(constants_1.INCOMPATIBLE_PEER_CODE, constants_1.INCOMPATIBLE_PEER_UNKNOWN_REASON);
                this.emit(events_2.EVENT_FAILED_TO_ADD_INBOUND_PEER, err);
                return;
            }
        };
        this._handleRemovePeer = (peerId) => {
            this.emit(events_2.EVENT_REMOVE_PEER, peerId);
        };
        this._handlePeerInfoUpdate = (peerInfo) => {
            if (!this._peerBook.hasPeer(peerInfo)) {
                this._peerBook.addPeer(peerInfo);
            }
            const isUpdated = this._peerBook.updatePeer(peerInfo);
            if (isUpdated) {
                this._peerBook.upgradePeer(peerInfo);
                this.emit(events_2.EVENT_UPDATED_PEER_INFO, peerInfo);
            }
        };
        this._handleFailedPeerInfoUpdate = (error) => {
            this.emit(events_2.EVENT_FAILED_PEER_INFO_UPDATE, error);
        };
        this._handleFailedToFetchPeerInfo = (error) => {
            this.emit(events_2.EVENT_FAILED_TO_FETCH_PEER_INFO, error);
        };
        this._handleFailedToFetchPeers = (error) => {
            this.emit(events_2.EVENT_FAILED_TO_FETCH_PEERS, error);
        };
        this._handleFailedToCollectPeerDetails = (error) => {
            this.emit(events_2.EVENT_FAILED_TO_COLLECT_PEER_DETAILS_ON_CONNECT, error);
        };
        this._handleBanPeer = (peerId) => {
            var _a;
            const banTime = (_a = this._config.peerBanTime) !== null && _a !== void 0 ? _a : constants_1.DEFAULT_BAN_TIME;
            if (this._peerPool.hasPeer(peerId)) {
                this._peerPool.removePeer(peerId);
            }
            this._peerBook.addBannedPeer(peerId, banTime);
            this.emit(events_2.EVENT_BAN_PEER, peerId);
        };
        this._handleDiscoveredPeer = (detailedPeerInfo) => {
            if (this._peerBook.hasPeer(detailedPeerInfo)) {
                return;
            }
            if (this._peerBook.addPeer(detailedPeerInfo)) {
                this.emit(events_2.EVENT_DISCOVERED_PEER, detailedPeerInfo);
            }
        };
        this._handleFailedToPushNodeInfo = (error) => {
            this.emit(events_2.EVENT_FAILED_TO_PUSH_NODE_INFO, error);
        };
        this._handleFailedToSendMessage = (error) => {
            this.emit(events_2.EVENT_FAILED_TO_SEND_MESSAGE, error);
        };
        this._handleOutboundSocketError = (error) => {
            this.emit(events_2.EVENT_OUTBOUND_SOCKET_ERROR, error);
        };
        this._handleInboundSocketError = (error) => {
            this.emit(events_2.EVENT_INBOUND_SOCKET_ERROR, error);
        };
        const peerPoolConfig = createPeerPoolConfig(config, this._peerBook);
        this._peerPool = new peer_pool_1.PeerPool(peerPoolConfig);
        this._bindHandlersToPeerPool(this._peerPool);
        this._nodeInfo = {
            ...config.nodeInfo,
            nonce: lisk_cryptography_1.getRandomBytes(constants_1.DEFAULT_NONCE_LENGTH_BYTES).toString('hex'),
        };
        this.applyNodeInfo(this._nodeInfo);
        this._populatorInterval = config.populatorInterval
            ? config.populatorInterval
            : constants_1.DEFAULT_POPULATOR_INTERVAL;
        this._fallbackSeedPeerDiscoveryInterval = config.fallbackSeedPeerDiscoveryInterval
            ? config.fallbackSeedPeerDiscoveryInterval
            : constants_1.DEFAULT_FALLBACK_SEED_PEER_DISCOVERY_INTERVAL;
        this._nextSeedPeerDiscovery =
            Date.now() + this._fallbackSeedPeerDiscoveryInterval;
    }
    get config() {
        return this._config;
    }
    get isActive() {
        return this._isActive;
    }
    applyNodeInfo(nodeInfo) {
        utils_1.validateNodeInfo(nodeInfo, this._config.maxPeerInfoSize
            ? this._config.maxPeerInfoSize
            : constants_1.DEFAULT_MAX_PEER_INFO_SIZE);
        this._nodeInfo = {
            ...nodeInfo,
            nonce: this.nodeInfo.nonce,
        };
        this._peerPool.applyNodeInfo(this._nodeInfo);
    }
    get nodeInfo() {
        return this._nodeInfo;
    }
    applyPenalty(peerPenalty) {
        this._peerPool.applyPenalty(peerPenalty);
    }
    getTriedPeers() {
        return this._peerBook.triedPeers.map(peer => ({
            ...peer.sharedState,
            ipAddress: peer.ipAddress,
            wsPort: peer.wsPort,
        }));
    }
    getConnectedPeers() {
        return this._peerPool
            .getAllConnectedPeerInfos()
            .filter(peer => !(peer.internalState && !peer.internalState.advertiseAddress))
            .map(peer => ({
            ...peer.sharedState,
            ipAddress: peer.ipAddress,
            wsPort: peer.wsPort,
            peerId: peer.peerId,
        }));
    }
    getDisconnectedPeers() {
        const allPeers = this._peerBook.allPeers;
        const connectedPeers = this.getConnectedPeers();
        const disconnectedPeers = allPeers.filter(peer => {
            if (connectedPeers.find(connectedPeer => peer.peerId === connectedPeer.peerId)) {
                return false;
            }
            return true;
        });
        return disconnectedPeers
            .filter(peer => !(peer.internalState && !peer.internalState.advertiseAddress))
            .map(peer => ({
            ...peer.sharedState,
            ipAddress: peer.ipAddress,
            wsPort: peer.wsPort,
            peerId: peer.peerId,
        }));
    }
    async request(packet) {
        const response = await this._peerPool.request(packet);
        return response;
    }
    send(message) {
        this._peerPool.send(message);
    }
    broadcast(message) {
        this._peerPool.broadcast(message);
    }
    async requestFromPeer(packet, peerId) {
        return this._peerPool.requestFromPeer(packet, peerId);
    }
    sendToPeer(message, peerId) {
        this._peerPool.sendToPeer(message, peerId);
    }
    _startPopulator() {
        if (this._populatorIntervalId) {
            throw new Error('Populator is already running');
        }
        this._populatorIntervalId = setInterval(() => {
            this._peerPool.triggerNewConnections(this._peerBook.newPeers, this._peerBook.triedPeers);
            if (this._nextSeedPeerDiscovery < Date.now() &&
                this._peerPool.getFreeOutboundSlots() > 0) {
                this._peerPool.discoverFromSeedPeers();
                this._nextSeedPeerDiscovery =
                    Date.now() + this._fallbackSeedPeerDiscoveryInterval;
            }
        }, this._populatorInterval);
        this._peerPool.triggerNewConnections(this._peerBook.newPeers, this._peerBook.triedPeers);
    }
    _stopPopulator() {
        if (this._populatorIntervalId) {
            clearInterval(this._populatorIntervalId);
        }
    }
    _isNetworkReady() {
        if (!this._hasConnected && this._peerPool.getConnectedPeers().length > 0) {
            this._hasConnected = true;
            return true;
        }
        return false;
    }
    _handleGetPeersRequest(request) {
        const minimumPeerDiscoveryThreshold = this._config
            .minimumPeerDiscoveryThreshold
            ? this._config.minimumPeerDiscoveryThreshold
            : constants_1.DEFAULT_MIN_PEER_DISCOVERY_THRESHOLD;
        const maxPeerDiscoveryResponseLength = this._config
            .maxPeerDiscoveryResponseLength
            ? this._config.maxPeerDiscoveryResponseLength
            : constants_1.DEFAULT_MAX_PEER_DISCOVERY_RESPONSE_LENGTH;
        const wsMaxPayload = this._config.wsMaxPayload
            ? this._config.wsMaxPayload
            : constants_1.DEFAULT_WS_MAX_PAYLOAD;
        const maxPeerInfoSize = this._config.maxPeerInfoSize
            ? this._config.maxPeerInfoSize
            : constants_1.DEFAULT_MAX_PEER_INFO_SIZE;
        const safeMaxPeerInfoLength = Math.floor(constants_1.DEFAULT_WS_MAX_PAYLOAD / maxPeerInfoSize) - 1;
        const selectedPeers = this._peerBook.getRandomizedPeerList(minimumPeerDiscoveryThreshold, maxPeerDiscoveryResponseLength);
        const sanitizedPeerInfoList = selectedPeers
            .filter(peer => !(peer.internalState && !peer.internalState.advertiseAddress))
            .map(peer => ({
            ipAddress: peer.ipAddress,
            wsPort: peer.wsPort,
            ...peer.sharedState,
        }));
        request.end({
            success: true,
            peers: utils_1.getByteSize(sanitizedPeerInfoList) < wsMaxPayload
                ? sanitizedPeerInfoList
                : sanitizedPeerInfoList.slice(0, safeMaxPeerInfoLength),
        });
    }
    _handleGetNodeInfo(request) {
        request.end(this._nodeInfo);
    }
    async start() {
        if (this._isActive) {
            throw new Error('Node cannot start because it is already active.');
        }
        if (this._config.maxInboundConnections !== 0) {
            this._peerServer = new peer_server_1.PeerServer({
                nodeInfo: this._nodeInfo,
                hostIp: this._config.hostIp || constants_1.DEFAULT_NODE_HOST_IP,
                secret: this._secret,
                peerBook: this._peerBook,
                maxPayload: this._config.wsMaxPayload
                    ? this._config.wsMaxPayload
                    : constants_1.DEFAULT_WS_MAX_PAYLOAD,
                maxPeerInfoSize: this._config.maxPeerInfoSize
                    ? this._config.maxPeerInfoSize
                    : constants_1.DEFAULT_MAX_PEER_INFO_SIZE,
                peerHandshakeCheck: this._config.peerHandshakeCheck
                    ? this._config.peerHandshakeCheck
                    : utils_1.validatePeerCompatibility,
            });
            this._bindHandlersToPeerServer(this._peerServer);
            try {
                await this._peerServer.start();
            }
            catch (err) {
                this._isActive = false;
                throw new Error('Peer server did not start successfully');
            }
        }
        this._isActive = true;
        if (this._isActive) {
            if (this._peerBook.triedPeers.length < constants_1.DEFAULT_MIN_TRIED_PEER_COUNT) {
                this._peerPool.discoverFromSeedPeers();
                this._nextSeedPeerDiscovery =
                    Date.now() + this._fallbackSeedPeerDiscoveryInterval;
            }
            this._startPopulator();
        }
    }
    async stop() {
        if (!this._isActive) {
            throw new Error('Node cannot be stopped because it is not active.');
        }
        this._isActive = false;
        this._hasConnected = false;
        this._stopPopulator();
        this._peerPool.removeAllPeers();
        this._peerBook.cleanUpTimers();
        if (this._peerServer) {
            await this._peerServer.stop();
            this._removeListeners(this._peerServer);
        }
        this._removeListeners(this._peerPool);
    }
    _bindHandlersToPeerPool(peerPool) {
        peerPool.on(events_2.EVENT_REQUEST_RECEIVED, this._handlePeerPoolRPC);
        peerPool.on(events_2.EVENT_MESSAGE_RECEIVED, this._handlePeerPoolMessage);
        peerPool.on(events_2.EVENT_CONNECT_OUTBOUND, this._handleOutboundPeerConnect);
        peerPool.on(events_2.EVENT_CONNECT_ABORT_OUTBOUND, this._handleOutboundPeerConnectAbort);
        peerPool.on(events_2.EVENT_CLOSE_INBOUND, this._handlePeerCloseInbound);
        peerPool.on(events_2.EVENT_CLOSE_OUTBOUND, this._handlePeerCloseOutbound);
        peerPool.on(events_2.EVENT_REMOVE_PEER, this._handleRemovePeer);
        peerPool.on(events_2.EVENT_UPDATED_PEER_INFO, this._handlePeerInfoUpdate);
        peerPool.on(events_2.EVENT_FAILED_PEER_INFO_UPDATE, this._handleFailedPeerInfoUpdate);
        peerPool.on(events_2.EVENT_FAILED_TO_FETCH_PEER_INFO, this._handleFailedToFetchPeerInfo);
        peerPool.on(events_2.EVENT_FAILED_TO_FETCH_PEERS, this._handleFailedToFetchPeers);
        peerPool.on(events_2.EVENT_FAILED_TO_COLLECT_PEER_DETAILS_ON_CONNECT, this._handleFailedToCollectPeerDetails);
        peerPool.on(events_2.EVENT_DISCOVERED_PEER, this._handleDiscoveredPeer);
        peerPool.on(events_2.EVENT_FAILED_TO_PUSH_NODE_INFO, this._handleFailedToPushNodeInfo);
        peerPool.on(events_2.EVENT_FAILED_TO_SEND_MESSAGE, this._handleFailedToSendMessage);
        peerPool.on(events_2.EVENT_OUTBOUND_SOCKET_ERROR, this._handleOutboundSocketError);
        peerPool.on(events_2.EVENT_INBOUND_SOCKET_ERROR, this._handleInboundSocketError);
        peerPool.on(events_2.EVENT_BAN_PEER, this._handleBanPeer);
    }
    _bindHandlersToPeerServer(peerServer) {
        peerServer.on(events_2.EVENT_BAN_PEER, this._handleBanPeer);
        peerServer.on(events_2.EVENT_INBOUND_SOCKET_ERROR, this._handleInboundSocketError);
        peerServer.on(events_2.EVENT_FAILED_TO_ADD_INBOUND_PEER, this._handleFailedInboundPeerConnect);
        peerServer.on(events_2.EVENT_NEW_INBOUND_PEER_CONNECTION, this._handleInboundPeerConnect);
    }
    _removeListeners(emitter) {
        emitter.eventNames().forEach((eventName) => {
            emitter.removeAllListeners(eventName);
        });
    }
}
exports.P2P = P2P;
//# sourceMappingURL=p2p.js.map